// spl.java

import java.util.*;

// ===============
// Class: language
// ===============

public class spl
{
   static boolean debug = false;

   // Variable memory space.
   static Vector v = new Vector (100);

   public static void main (String [] args)
   {
      // Program being interpreted.
      String [] program =
      {
         "var j",
         "j = 13 - 2",
         "var i",
         "i = ( j + 3 ) * 5",
         "print",
         "x = 15"
      };

      // For each program line...
      for (int i = 0; i < program.length; i++)
      {
           // Create a tokenizer for the current program line.
           Tokenizer tok =
                     new Tokenizer (new StringTokenizer
                                                   (program [i]));

           // Fetch the first token from the String.
           String token = tok.getToken (false);

           if (debug)
               System.out.println ("\nNEW LINE!  First token is: " 
                                   + "[" + token + "]\n");

           // If line is blank -- indicated by "" -- then continue
           // to the next line.
           if (token.equals (""))
               continue;

           // Parse the current line.
           try
           {
              // If the line starts with "var" keyword then
              // create a variable.
              if (token.equalsIgnoreCase ("var"))
                  create_variable (tok);

              else
              // If line starts with "print" keyword then
              // print all existing variables names and values.
              if (token.equalsIgnoreCase ("print"))
                  print_variables (tok);

              else
              // By default, line must be an assignment statement.
                  assign_expr_to_variable (tok);
           }
           catch (SyntaxException e)
           {
              // Print out a syntax error message.
              System.out.println (e.getMessage ());

              // Terminate the program.
              return;
           }
      }
   }

   static void create_variable (Tokenizer tok)
               throws SyntaxException
   {
      if (debug)
          System.out.println ("Entering create_variable...");

      // Get token containing variable name from the cache.
      String varname = tok.getToken (false);

      if (debug)
          System.out.println ("Variable name = [" + varname +
                              "]");

      // If end-of-line reached, there is no variable name.
      if (varname == null)
          throw new SyntaxException ("missing variable name");

      // Make sure that varname is an identifier.
      if (!validIdentifier (varname))
          throw new SyntaxException (varname + " is not a valid"
                                    + " identifier.");

      // Create a new Variable object of integer type whose name
      // is specified by varname and whose initial value is 0.
      v.addElement (new Variable (varname, 0));

      if (debug)
          System.out.println ("Leaving create_variable...");
   }

   static void print_variables (Tokenizer tok)
   {
      if (debug)
          System.out.println ("Entering print_variables...");

      // Obtain an enumeration object from the Vector object v.
      Enumeration e = v.elements ();

      // Iterate through all Variable objects printing out names 
      // and values.
      while (e.hasMoreElements ())
         System.out.println (e.nextElement ());

      if (debug)
          System.out.println ("Leaving print_variables...");
   }

   static void assign_expr_to_variable (Tokenizer tok)
               throws SyntaxException
   {
      if (debug)
          System.out.println ("Entering " +
                              "assign_expr_to_variable...");

      // Get the variable name from the cache.
      String varname = tok.getToken (true);

      // Make sure that varname is an identifier.
      if (!validIdentifier (varname))
          throw new SyntaxException (varname +
                                     " is not a valid " +
                                     "identifier.");

      // Make sure that a Variable object exists whose name
      // matches varname.
      Variable var = findVariable (varname);

      if (var == null)
          throw new SyntaxException (varname +
                                     " is not defined!");

      // Get the next token from the String.
      String token = tok.getToken (false);

      if (debug)
          System.out.println ("Token after variable name = [" +
                              token + "]");

      // The next token must be an "=" assignment operator.
      if (!token.equals ("="))
          throw new SyntaxException ("Missing = character.");

      // Parse the expression after the "=" operator.
      int value = parse_expression (tok);

      if (debug)
          System.out.println ("Last token processed by " +
                              "parse_expression " +
                              " and still in cache = [ " +
                              tok.getToken (true) + "]");

      // Store the expression's value in the variable.
      var.setValue (value);

      if (debug)
          System.out.println ("Leaving assign_expr_to_variable" +
                              "...");
   }

   static boolean validIdentifier (String identifier)
   {
      // Assume success.

      boolean status = true;

      if (debug)
          System.out.println ("Entering validIdentifer with [" +
                              identifier + "]...");

      loop:
      do
      {
          // Safety checking (to prevent exception).

          if (identifier == null || identifier.length () == 0)
          {
              status = false;
              break;
          }

          if (!Character.isJavaIdentifierStart (
                                           identifier.charAt (0)))
          {
              status = false;
              break;
          }

          for (int i = 1; i < identifier.length (); i++)
               if (!Character.isJavaIdentifierPart (
                                           identifier.charAt (i)))
               {
                   status = false;
                   break loop;
               }
      }
      while (false);

      if (debug)
          System.out.println ("Leaving validIdentifer...");

      return status;
   }

   static Variable findVariable (String name)
   {
      // Assume variable not found.

      Variable var = null;

      if (debug)
          System.out.println ("Entering findVariable with [" +
                              name + "]...");

      // Obtain an enumeration object from the Vector object v.
      Enumeration e = v.elements ();

      // Iterate through all Variable objects until object found
      // or not.
      while (e.hasMoreElements ())
      {
         Variable v = (Variable) e.nextElement ();
         if (v.getName ().equals (name))
         {
             var = v;
             break;
         }
      }

      if (debug)
          System.out.println ("Leaving findVariable...");

      return var;
   }

   static int parse_expression (Tokenizer tok)
              throws SyntaxException
   {
      if (debug)
          System.out.println ("Entering parse_expression...");

      int value = parse_term (tok);

      String token = tok.getToken (true);

      if (debug)
          System.out.println ("First expression token = [" + token
                              + "]");

      if (token.equals ("+"))
          value += parse_term (tok);
      else
      if (token.equals ("-"))
          value -= parse_term (tok);

      if (debug)
          System.out.println ("Leaving parse_expression...");

      return value;
   }

   static int parse_term (Tokenizer tok) throws SyntaxException
   {
      if (debug)
          System.out.println ("Entering parse_term...");

      int value = parse_factor (tok);

      String token = tok.getToken (true);

      if (debug)
          System.out.println ("First term token = [" + token +
                              "]");

      if (token.equals ("*"))
          value *= parse_factor (tok);
      else
      if (token.equals ("/"))
          value /= parse_factor (tok);

      if (debug)
          System.out.println ("Leaving parse_term...");

      return value;
   }

   static int parse_factor (Tokenizer tok) throws SyntaxException
   {
      if (debug)
          System.out.println ("Entering parse_factor...");

      String factor = tok.getToken (false);
      if (factor.equals (""))
          throw new SyntaxException ("factor expected");

      if (debug)
          System.out.println ("First factor token = [" + factor +
                              "]");

      int value;

      if (validIdentifier (factor))
      {
          Variable var = findVariable (factor);
          if (var == null)
              throw new SyntaxException ("Variable " + factor +
                                         " not defined.");

          value = var.getValue ();
      }
      else
      if (factor.equals ("("))
      {
          if (debug)
              System.out.println ("Recursively calling " +
                                  "parse_expression...");

          value = parse_expression (tok);

          if (debug)
              System.out.println ("Leaving recursed " +
                                  "parse_expression...");

          if (debug)
              System.out.println ("Last token processed by " +
                                  "parse_expression " +
                                  "and still in cache = [ " +
                                  tok.getToken (true) + "]");
        
          if (!tok.getToken (true).equals (")"))
              throw new SyntaxException ("Closing bracket " +
                                         " expected.");
      }
      else
      {
          for (int i = 0; i < factor.length (); i++)
               if (!Character.isDigit (factor.charAt (i)))
                   throw new SyntaxException ("Digit expected");

          value = Integer.parseInt (factor);
      }
      
      // Get next token after factor.  Token is cached.
      tok.getToken (false);

      if (debug)
          System.out.println ("Leaving parse_factor...");

      return value;
   }
}

// ======================
// Class: SyntaxException
// ======================

class SyntaxException extends Exception
{
   SyntaxException (String msg)
   {
      super (msg);
   }
}

// ================
// Class: Tokenizer
// ================

class Tokenizer
{
   private StringTokenizer st;
   private String token = "";

   Tokenizer (StringTokenizer st)
   {
      this.st = st;
   }

   String getToken (boolean getCachedToken)
   {
      if (getCachedToken)
          return token;

      try
      {
         token = (String) st.nextToken ();
      }
      catch (NoSuchElementException e)
      {
         token = "";
      }

      return token;
   }
}

// ===============
// Class: Variable
// ===============

class Variable
{
   private String name;
   private int value;

   Variable (String name, int value)
   {
      this.name = name;
      setValue (value);
   }

   String getName () { return name; }

   int getValue () { return value; }
   void setValue (int value) { this.value = value; }

   public String toString () { return name + " " + value; }
}
